`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company:         BME-VIK
// Engineer:        Peter Czako
// 
// Create Date:     16:52:35 01/21/2007  
// Design Name: 
// Module Name:     bp_logic 
// Project Name:    XSOC
// Target Devices:  SPARTAN-3, SPARTAN-3E
// Tool versions:   XILINX ISE 8.2
// Description:     Break point logic for the XSOC system. A 16 bit value can
//                  be set in the match register of the breakpoint module, and
//                  if this value occurs on the address bus, an interrupt is 
//                  generated.
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////

//=========================================================
// MODULE DECLARATION
//=========================================================

module bp_logic(
// XSOC BUS INTERFACE
    input           clk,
    input           rst,
    input           rnw,
    input   [1:0]   be,
    input   [15:0]  addr,
    input   [15:0]  din,
    output  [15:0]  dout,
    output  reg     irq,
    input           irq_served
// USER IO PORTS    
    );
    
//========================================================= 
// XSOC BUS INTERFACE LOGIC 
//=========================================================
    parameter   BASE_ADDRESS = 16'h6004;
    
    reg [15:0]  to_dout;
    
    wire        sel;

    assign sel          = addr[15:2] == BASE_ADDRESS >> 2;     
    assign dout[7:0]    = ~clk & sel & rnw & be[0] ? to_dout[7:0] : 8'b0;
    assign dout[15:8]   = ~clk & sel & rnw & be[1] ? to_dout[15:8] : 8'b0;

//========================================================= 
// USER LOGIC
//=========================================================

// REGS
    reg         irq_en;                 //  Interrupt enable
    reg [15:0]  match_reg;              // Address to match with
    
// WIRES
    wire        bp_match;               // Match occured

// ASSIGNS  
    assign bp_match = irq_en & (addr == match_reg);

//========================================================= 
// WRITE REGISTERS
//=========================================================
    always @(negedge clk) begin
        if (rst) begin 
            match_reg <= 0;
            irq_en <= 0;
        end
        else if (sel & ~rnw & ~addr[1]) begin
            if (be[0])
                match_reg[7:0] <= din[7:0];
            if (be[1])
                match_reg[15:8] <= din[15:8];
        end
        else if (sel & ~rnw & addr[1] & be[1])
            irq_en <= din[8];
    end

//=========================================================     
// READ REGISTERS
//=========================================================
    always @(negedge clk) begin
        if (rst)
            to_dout <= 0;
        else if (sel & rnw & ~addr[1])
            to_dout <= match_reg;
        else if (sel & rnw & addr[1])
            to_dout <= {7'b0,irq_en,8'b0};
        else
            to_dout <= 0;
    end

//=========================================================
// INTERRUPT REQUEST HANDLING
//=========================================================
    always @ (negedge clk) begin
        if (rst)
            irq <= 0;
        else if (bp_match)
            irq <= bp_match;
        else if (irq_served)
            irq <= 0;
    end
    
endmodule
