`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company:         BME-VIK
// Engineer:        Peter Czako
// 
// Create Date:     11/25/2006  
// Design Name: 
// Module Name:     ctrl 
// Project Name:    XSOC
// Target Devices:  SPARTAN-3, SPARTAN-3E
// Tool versions:   XILINX ISE 8.2
// Description:     The control submodule of the xr16 processor.
//                  Based on the original design of Jan Gray. 
//                  Modifications implemented: 
//                  - New operations (MUL, REG RELATIVE BARREL SHIFT)
//                  - Interrupt handling corrected
//                  - Result multiplexer select signal simplified
//                  - Flag setting operations made interruptable
//                  - DMA handling reconstructed
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: Used for educational purposes only
//
//////////////////////////////////////////////////////////////////////////////////

/* ctrl.v -- xr16 control unit synthesizable Verilog model
 *
 * Copyright (C) 1999, 2000, Gray Research LLC.  All rights reserved.
 * The contents of this file are subject to the XSOC License Agreement;
 * you may not use this file except in compliance with this Agreement.
 * See the LICENSE file.
 *
 * $Header: /dist/xsocv/ctrl.v 8     4/06/00 10:55a Jan $
 * $Log: /dist/xsocv/ctrl.v $
 * 
 * 8     4/06/00 10:55a Jan
 * polish
 */


// Instruction fields, ref.: The xr16 Specifications (doc/xspecs.pdf).

// Opcodes (IR[15:12])
`define ADD     4'h0
`define SUB     4'h1
`define ADDI    4'h2
`define RR      4'h3
`define RI      4'h4
`define LW      4'h5
`define LB      4'h6
`define SW      4'h8
`define SB      4'h9
`define JAL     4'hA
`define Bcond   4'hB
`define CALL    4'hC
`define IMM     4'hD
`define MUL     4'hE

// Functions (IR[7:4])
`define AND     4'h0
`define OR      4'h1
`define XOR     4'h2
`define ANDN    4'h3
`define ADC     4'h4
`define SBC     4'h5
`define SRL     4'h6
`define SRA     4'h7
`define SLL     4'h8

// Conditional branches (IR[11:8])
`define BR      4'h0
`define BRN     4'h1
`define BEQ     4'h2
`define BNE     4'h3
`define BC      4'h4
`define BNC     4'h5
`define BV      4'h6
`define BNV     4'h7
`define BLT     4'h8
`define BGE     4'h9
`define BLE     4'hA
`define BGT     4'hB
`define BLTU    4'hC
`define BGEU    4'hD
`define BLEU    4'hE
`define BGTU    4'hF

`define INT     16'hAE01    // int ::= jal r14,10(r0)
`define RETI    16'hA0E0    // reti :: = jal r0,0(r14)

//========================================================= 
// MODULE DECLARATION
//========================================================= 

module control(
    input               clk,            // global clock
    input               rst,            // global async reset
    input               nwait,          // stops the processor to execute operations
    input               irq,            // interrupt request
    input               dma_req,        // direct memory access request
    input       [15:0]  insn,           // new instruction word
    input               z,              // zero result condition code
    input               n,              // negative result condition code
    input               co,             // carry-out result codition code
    input               v,              // oVerflow result condition code
    output  reg         rnw,            // next access is read
    output  reg         dma_ack,        // direct memory access acknowledged
    output  reg         irq_served,     // interrupt service done
    output              rdy,            // ready to operate
    output  reg         word,           // next access is word wide
    output              rf_we,          // register file write enable
    output      [3:0]   rna,            // register file port A register number
    output      [3:0]   rnb,            // register file port B register number
    output              fwd,            // forward result bus into A operand register
    output      [11:0]  imm,            // 12-bit immediate field
    output              sextimm4,       // sign-extend 4-bit immediate operand
    output              zextimm4,       // zero-extend 4-bit immediate operand
    output              wordimm4,       // word-offset 4-bit immediate operand
    output              imm12,          // 12-bit immediate operand
    output              pipe_ce,        // pipeline clock enable
    output              b15_4_ce,       // b[15:4] clock enable
    output  reg         add,            // 1 => A + B; 0 => A - B
    output  reg         ci,             // carry-in
    output      [1:0]   logicop,        // logic unit opcode
    output              sll,            // shift left logical operand
    output              srl,            // shift right logical operand
    output              sra,            // shift right arithmetical operand
    output  reg [2:0]   res_sel,        // result select signals
    output  reg         sync_reset,     // 1 cycle delayed reset
    output  reg         branch,         // branch taken
    output      [7:0]   brdisp,         // 8-bit branch displacement
    output              selpc,          // address mux selects next PC
    output              pc_ce,          // PC clock enable
    output              ret_ce          // return address clock enable
    );
    
//========================================================= 
// LOCALS
//========================================================= 

// PARAMETERS
// Result multiplexer select values
    parameter SEL_EXTERNAL          = 3'b000;
    parameter SEL_SUM               = 3'b001;
    parameter SEL_LOGIC             = 3'b010;
    parameter SEL_RIGHT_SH          = 3'b011;
    parameter SEL_MUL_LEFT_SH       = 3'b100;
    parameter SEL_RET               = 3'b101;

// REGS
    reg [15:0]  if_ir;          // IF stage instruction reg
    reg [15:0]  ir;             // DC stage instruction reg
    reg [15:0]  ex_ir;          // EX stage instruction reg
    reg         ex_call;        // EX stage call
    reg         ex_st;          // EX stage store
    reg         ifetch;         // "current access is insn fetch"
    reg         int_pend;       // interrupt is pending
    reg         dc_int;         // DC stage instruction is int
    reg         it_taken;       // interrupt is under service
    reg         dc_annul;       // annul DC stage instruction
    reg         ex_annul;       // annul EX stage instruction
    reg         z_reg;          // zero flag saved value    
    reg         n_reg;          // negative flag saved value    
    reg         co_reg;         // carry out flag saved value
    reg         v_reg;          // overflow flag saved value
    reg         irq_served_1d;  // interrupt served 1 cycle delayed
    
// WIRES
    // IR fields
    wire [3:0]  op     = ir[15:12];        // opcode
    wire [3:0]  rd     = ir[11:8];         // destination register
    wire [3:0]  ra     = ir[7:4];          // source register A
    wire [3:0]  rb     = ir[3:0];          // source register B
    wire [3:0]  cond   = ir[11:8];         // branch condition
    wire [3:0]  fn     = ir[7:4];          // rr or ri function sub-opcode
    wire [3:0]  ex_op  = ex_ir[15:12];     // EX stage opcode
    wire [3:0]  ex_rd  = ex_ir[11:8];      // EX stage destination register
    assign      imm    = ir[11:0];         // 12-bit immediate field
    assign      brdisp = ex_ir[7:0];       // 8-bit branch displacement

    wire        if_int;         // IF stage 'interrupt in progress'

    // IF stage instruction decoding
    wire        addsub  = op==`ADD || op==`SUB || op==`ADDI;
    wire        call    = op==`CALL;
    wire        st      = op==`SW || op==`SB;
    wire        rrri    = op==`RR || op == `RI;
    wire        adcsbc  = rrri && (fn==`ADC || fn==`SBC);
    wire        dcintinh= op==`Bcond || op==`CALL || op==`IMM || op==`JAL; // not interruptable insrtuctions
    wire        sub     = op==`SUB || (rrri && (fn==`SBC));
    
    assign imm12        = op==`CALL || op==`IMM;
    assign sextimm4     = op==`ADDI || (op==`RI && !(fn==`SRL || fn==`SRA || fn==`SLL));
    assign zextimm4     = op==`LB || op==`SB || (op==`RI && (fn==`SRL || fn==`SRA || fn==`SLL));
    assign wordimm4     = op==`LW || op==`SW || op==`JAL;
    assign sll          = rrri && (fn==`SLL);
    assign sra          = rrri && (fn==`SRA);
    assign srl          = rrri && (fn==`SRL);
    assign rdy          = nwait & ~dma_ack;     // operation stops in case of DMA 
    
    // EX stage decoding
    wire ex_ldst        = ex_op== `LW || ex_op==`LB || ex_op==`SW || ex_op==`SB;
    wire ex_lbsb        = ex_op==`LB || ex_op==`SB;
    wire ex_results     = ex_op==`ADD || ex_op==`SUB || ex_op==`ADDI || ex_op==`MUL ||
                          ex_op==`RR || ex_op==`RI ||
                          ex_op==`LW || ex_op==`LB || ex_op==`JAL || ex_op==`CALL;
    wire ex_jump        = ex_op==`JAL || ex_op==`CALL;
    
   
// memory access FSM transitions:
// case (state)
//   IF: state_nxt = ldst_pend ? LS : IF;
//   LS: state_nxt = IF;
// endcase
    wire ldst_pend      = ex_ldst & ~ex_annul;
    wire if_nxt         = ifetch&~ldst_pend | ~ifetch;
    wire ldst_nxt       = ifetch&ldst_pend;
    wire jump           = ex_jump && ~ex_annul;
    wire exannul_nxt    = sync_reset | branch | jump | dc_annul;

// FSM output decodes
    wire word_nxt       = ~(ldst_nxt&ex_lbsb);
    wire read_nxt       = ~(ldst_nxt&ex_st);
    assign selpc        = if_nxt&~jump ;
    assign pipe_ce      = rdy&if_nxt;
    assign pc_ce        = rdy&(if_nxt);
    assign ret_ce       = rdy&if_nxt&~dc_int;
    
// Interrupt handling signals
    

    // interrupt is disabled during interrupt service
    // Preemptive interrupts are not available
    // If the interrupt service is done, the flags must be reloaded
    wire z_int          = irq_served_1d ? z_reg : z;
    wire n_int          = irq_served_1d ? n_reg : n;
    wire co_int         = irq_served_1d ? co_reg : co;
    wire v_int          = irq_served_1d ? v_reg : v;
    wire it_en          = ~branch & ~jump & ~dcintinh & ~it_taken ;//& pipe_ce;
    assign if_int       = int_pend & it_en;



//=========================================================
// INTERRUPT HANDLING
//=========================================================

// Saving the flags if interrupt occured
    always @ (posedge clk) begin
        if (rst) begin
            z_reg   <= 0;
            n_reg   <= 0;
            co_reg  <= 0;
            v_reg   <= 0;
        end
        else if (dc_int) begin
            z_reg   <= z;
            n_reg   <= n;
            co_reg  <= co;
            v_reg   <= v;
        end
    end

// Generating interrupt served signals    
    always @ (posedge clk) begin
        if (rst) begin
            irq_served <= 0;
            irq_served_1d <= 0;
        end
        else begin
            irq_served_1d <= irq_served;
            if (ex_ir == `RETI) 
                irq_served <= 1;
            else
                irq_served <= 0;
        end
    end

// it_taken is active from irq accept until RETI is executed
    always @ (posedge clk) begin
        if (rst)
            it_taken <= 0;
        else if (if_int & pipe_ce) 
            it_taken <= 1;
        else if (ex_ir == `RETI) 
            it_taken <= 0;
    end

// interrupt request clocking
    always @(posedge clk) begin
        if (rst | (if_int & pipe_ce) )
            int_pend <= 0;
        else if (irq & ~it_taken)
            int_pend <= 1;
    end
    
    always @(posedge clk) begin
        if (rst)
            dc_int <= 0;
        else if (pipe_ce)
            dc_int <= if_int;
    end



//=========================================================
// OPERATIONAL STATES : INSTRUCTION FETCH, DECODE, EXECUTE
//=========================================================

// INSTRUCTION FETCH

// IR clocking
    always @(posedge clk ) begin
        if (rst) begin
            if_ir <= 0;
            ir <= 0;
            ex_ir <= 0;
        end
        else begin
            if (ifetch && rdy)
                if_ir <= insn;
            if (pipe_ce) begin
                ir <= (if_int) ? `INT : (ifetch) ? insn : if_ir;
                ex_ir <= ir;
            end
        end
    end

// instruction decode clocking
    always @(posedge clk ) begin
        if (rst) begin
            ex_call <= 0;
            ex_st <= 0;
        end
        else if (pipe_ce) begin
            ex_call <= call;
            ex_st <= st;
        end
    end

// memory access FSM clocking
    always @(posedge clk ) begin
        if (rst) begin
            ifetch <= 1;
        end
        else if (rdy) begin
            ifetch <= if_nxt;
        end
    end

// annul states clocking
    always @(posedge clk ) begin
        if (rst) begin
            dc_annul <= 1;
            ex_annul <= 1;
        end
        else if (pipe_ce) begin
            dc_annul <= sync_reset | branch | jump;
            ex_annul <= exannul_nxt;
        end
    end


// DECODE STAGE


// DC stage operand selection
    wire [3:0]  rsrc    = call ? 0 : rrri ? rd : ra;
    wire [3:0]  rdest   = ex_call ? 15 : ex_rd;
    assign      rna     = clk ? rsrc : rdest;
    assign      rnb     = clk ? (st ? rd : rb) : (ex_call ? 15 : ex_rd);
    assign      rf_we   = ex_results & pipe_ce & ~ex_annul & (rdest != 0);
    assign      fwd     = (rsrc==rdest) & (rdest!=0) && ex_results & ~ex_annul;
    assign      b15_4_ce= pipe_ce && ~(ex_op==`IMM && ~ex_annul);


// DC stage conditional branches
    reg taken;
    always @(posedge clk ) begin
        if (rst)
            branch = 0;
        else if (pipe_ce) begin
            case (cond)
            `BR:   taken = 1;
            `BRN:  taken = 0;
            `BEQ:  taken = z_int;
            `BNE:  taken = ~z_int;
            `BC:   taken = co_int;
            `BNC:  taken = ~co_int;
            `BV:   taken = v_int;
            `BNV:  taken = ~v_int;
            `BLT:  taken = n_int^v_int;
            `BGE:  taken = ~(n_int^v_int);
            `BLE:  taken = (n_int^v_int)|z_int;
            `BGT:  taken = ~((n_int^v_int)|z_int);
            `BLTU: taken = ~z_int&~co_int;
            `BGEU: taken = z_int|co_int;
            `BLEU: taken = z_int|~co_int;
            `BGTU: taken = ~z_int&co_int;
            endcase
            branch = op==`Bcond && taken && ~exannul_nxt;
        end
    end


// EXECUTION STAGE

// EX stage ALU control
    always @(posedge clk ) begin
        if (rst) begin
            add <= 1;
            ci <= 0;
        end
        else if (pipe_ce) begin
            add <= ~sub;
            ci <= sub ^ (adcsbc & co);
        end
    end
    assign  logicop = ex_ir[5:4];

// EX stage result multiplexer control
    always @(posedge clk ) begin
        if (rst) 
            res_sel <= 0;
        else if (pipe_ce) begin
            if (addsub || adcsbc)
                res_sel <= SEL_SUM;       
            else if (rrri&& (fn==`AND || fn==`OR || fn==`XOR || fn==`ANDN))
                res_sel <= SEL_LOGIC;       
            else if(sra | srl )
                res_sel <= SEL_RIGHT_SH;      
            else if ( sll || op==`MUL)
                res_sel <= SEL_MUL_LEFT_SH;      
            else if (op==`JAL || op==`CALL)
                res_sel <= SEL_RET;       
            else
                res_sel <= SEL_EXTERNAL;      
        end
    end

//=========================================================
// OTHER 
//=========================================================

// RESET CLOCKING
    always @(posedge clk ) begin
        if (rst)
            sync_reset <= 1;
        else if (rdy)
            sync_reset <= 0;
    end

// DMA HANDLING
    always @ (posedge clk) begin
        if (rst)
            dma_ack <=0;
        else if (nwait)
            dma_ack <= dma_req;
    end

// WORD AND RNW SIGNAL CLOCKING
    always @(posedge clk) begin
        if (rst) begin
            word <= 1;
            rnw <= 1;
        end
        else if (rdy) begin
            word <= word_nxt;
            rnw <= read_nxt;
        end
    end

endmodule
