`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company:         BME-VIK
// Engineer:        Peter Czako
// 
// Create Date:     20:26:32 04/23/2007  
// Design Name: 
// Module Name:     datapath 
// Project Name:    XSOC
// Target Devices:  SPARTAN-3, SPARTAN-3E
// Tool versions:   XILINX ISE 8.2
// Description:     The datapath submodule of the xr16 processor.
//                  Based on the original design of Jan Gray. 
//                  Modifications implemented: 
//                  - Async Multiplier implemented (MUL, BARREL SHIFT)
//                  - Result multiplexer select signal simplified
//                  - Data bus separated to in and out directions
//                  - Tri-state buffers removed
//                  - B operand preparation upgraded
//                  - Byte enable signals implemented
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: Used for educational purposes only
//
//////////////////////////////////////////////////////////////////////////////////

/* datapath.v -- xr16 datapath synthesizable Verilog model
 *
 * Copyright (C) 1999, 2000, Gray Research LLC.  All rights reserved.
 * The contents of this file are subject to the XSOC License Agreement;
 * you may not use this file except in compliance with this Agreement.
 * See the LICENSE file.
 *
 * $Header: /dist/xsocv/datapath.v 6     4/06/00 10:55a Jan $
 * $Log: /dist/xsocv/datapath.v $
 * 
 * 6     4/06/00 10:55a Jan
 * polish
 */ 


//=========================================================
// MODULE DECLARATION
//=========================================================

module datapath(
    input               clk,            // global clock
    input               rst,            // global async reset
    input               rf_we,          // register file write enable
    input       [3:0]   rna,            // register file port A register number
    input       [3:0]   rnb,            // register file port B register number
    input               fwd,            // forward result bus into A operand reg
    input       [11:0]  imm,            // 12-bit immediate field
    input               sextimm4,       // sign-extend 4-bit immediate operand
    input               zextimm4,       // zero-extend 4-bit immediate operand
    input               wordimm4,       // word-offset 4-bit immediate operand
    input               imm12,          // 12-bit immediate operand
    input               pipe_ce,        // pipeline clock enable
    input               b15_4_ce,       // b[15:4] clock enable
    input               add,            // 1 => A + B, 0 => A - B
    input               ci,             // carry-in
    input       [1:0]   logicop,        // logic unit opcode
    input       [2:0]   res_sel,        // kimenet kivalaszto jel
    input               sll,            // logikai balra shift jelzese
    input               sra,            // aritmetikai jobbra shift jelzese
    input               srl,            // logikai jobbra shift jelzese
    input               sync_reset,     // 1 cycle delayed reset
    input               branch,         // branch taken
    input       [7:0]   brdisp,         // 8-bit branch displacement
    input               selpc,          // address mux selects next PC
    input               pc_ce,          // PC clock enable
    input               ret_ce,         // return address clock enable
    input               word,           // next access is word wide
    input               rdy,            // ready to operate
    input       [15:0]  data_in,        // 16 bit data in bus
    output  reg         z,              // zero result condition code
    output  reg         n,              // negative result condition code
    output  reg         co,             // carry-out result codition code
    output  reg         v,              // oVerflow result condition code
    output  reg [15:0]  addr,           // address of next memory access
    output      [1:0]   be,             // byte enable
    output      [15:0]  data_out        // 16 bit data out bus
    );
    
//=========================================================
// LOCALS
//=========================================================

// PARAMETERS
// Result multiplexer select values
    parameter SEL_EXTERNAL          = 3'b000;
    parameter SEL_SUM               = 3'b001;
    parameter SEL_LOGIC             = 3'b010;
    parameter SEL_RIGHT_SH          = 3'b011;
    parameter SEL_MUL_LEFT_SH       = 3'b100;
    parameter SEL_RET               = 3'b101;

    reg  [15:0] res;                    // internal result
    reg  [15:0] addr_nxt;               // next address value
    
    wire [15:0] areg_nxt, breg_nxt;     // reg file read port RAM output buses
    reg  [15:0] areg, breg;             // register file read ports
    reg  [15:0] a, b;                   // operand registers
    reg  [15:0] dout;                   // store data output register
    reg  [15:0] sum;                    // adder output
    reg  [15:0] logic;                  // logic unit output
    reg  [15:0] pc;                     // current PC
    reg  [15:0] pcnext;                 // next PC  // REVIEW
    reg  [15:0] ret;                    // return address (DC stage PC)
    wire [6:0]  brdispext = {brdisp[7],brdisp[7],brdisp[7],brdisp[7],
                                 brdisp[7],brdisp[7],brdisp[7]};
//=========================================================
// SUBMODULES
//=========================================================

// 2 x 16 x 16 bit register file
    ram16x16s aregs(.wclk(clk), .addr(rna), .we(rf_we), .d(res), .o(areg_nxt));
    ram16x16s bregs(.wclk(clk), .addr(rnb), .we(rf_we), .d(res), .o(breg_nxt));


//=========================================================
// SIGNALS FOR THE MODIFICATIONS
//=========================================================

    reg  [15:0] to_b;                   // signal after IMM logic to B operand
    wire [15:0] op_b;                   // signal after Shift logic to B operand 
    reg  [15:0] sra_sign_ext;           // arithmetic right shift sign extension
    wire [15:0] to_sra_sign_ext;        // signal before sra_sign_ext
    wire [3:0]  to_dec;                 // shift ammount before the decoder
    wire [15:0] to_sh_b;                // shift ammount after decoding
    wire [15:0] sh_b;                   // shift ammount to multiplier
    wire [15:0] op_a;                   // A operand wires
    wire        to_b16;                 // signal before b16
    reg         b16;                    // used only if shift ammount is 0
    wire        sh_err;                 // shift operation error
    reg  [31:0] mul;                    // multiplier result

// shift result is in the upper 16 bit of the multiplier.
// sign extension used through an OR gate 
    wire [15:0] shr_res = (mul[31:16] | sra_sign_ext[15:0]);
        
// ALU signals
    reg         ign;    // not used, only for coding purposes
    reg         co0;    // carry out temp value

//=========================================================
// BARREL SHIFT WITH MULTIPLIER
//=========================================================

// if right shift, 2's complement on to_b[3:0]
    assign to_dec = (sll & ~(srl | sra)) ? to_b[3:0] : -to_b[3:0];
    
// 2^ammount goes to the shift operand
    assign to_sh_b = 16'h0001 << to_dec;

// Shift error if B operand is larger than 15. 
    assign sh_err = |(to_b[15:4]);    

// Sign extension submodule
    sign_mux mux_sign(
                    .sign(op_a[15]),
                    .error(sh_err),
                    .arit(sra),
                    .ammount(to_b[3:0]),
                    .res(to_sra_sign_ext)
                    );

// Set b16 if good right shift operation executed, and shift ammount is 0
    assign to_b16       = ~sh_err & (srl | sra) & (to_b[3:0] == 0);
// Set shift operand if no error occured     
    assign sh_b[15:1]   = ~sh_err ? to_sh_b[15:1] : 15'b0;
// Set shift operand lowest bit to 0 if error occured OR b16 is set
    assign sh_b[0]      = ~sh_err & ~to_b16 ? to_sh_b[0] : 1'b0;

//=========================================================
// OPERAND PRELOAD
//=========================================================

// A operand forwarding
    assign op_a = fwd ? res : areg; 
 
// B operand is filled with the shift ammount in case of shift operation
    assign op_b = (sll | srl | sra) ? sh_b : to_b;
    

//=========================================================
// BYTE ENABLE
//=========================================================
    assign be[0] = word ^ addr[0];
    assign be[1] = ~addr[0];    

//=========================================================
// DATA INPUT AND OUTPUT HANDLING
//=========================================================
  
    wire [15:0] external;       // external data read from the bus

// BYTE READ FROM MSB -> INPUT MSB PUT INTO LSB REGISTER AND ZERO EXTENSION ON MSB
    assign external[7:0]  = (be[1] & ~be[0]) ? data_in[15:8] : data_in[7:0];
    assign external[15:8] = (be[1] & be[0])  ? data_in[15:8] : 8'b0;
    
// BYTE WRITE TO MSB -> OUTPUT LSB PUT INTO MSB REGISTER AS WELL
    assign data_out[15:8] = (be[1] & ~be[0]) ? dout[7:0] : dout[15:8];
    assign data_out[7:0]  = dout[7:0];


//=========================================================
// PC clocking
//=========================================================
    always @ (posedge clk) begin
        if (rst)
            pc <= 0;
        else if (pc_ce)
            pc <= addr_nxt;
    end
    
//=========================================================
// register file output registers
//=========================================================
    always @(negedge clk ) begin
        if (rst) begin
            areg <= 0;
            breg <= 0;
        end
        else begin
            areg <= areg_nxt;
            breg <= breg_nxt;
        end
    end

//=========================================================
// B operand immediate preload
//=========================================================
    always @(sextimm4,zextimm4,wordimm4,imm12,imm,breg) begin
        if (sextimm4 || zextimm4) begin
            to_b[3:0] = imm[3:0];
            if (sextimm4)
                to_b[15:4] = {12{imm[3]}};
            else to_b[15:4] = 12'b0;
        end
        else if (wordimm4)
            to_b = {11'b0,imm[0],imm[3:1],1'b0};
        else if (imm12)
            to_b = {imm[11:0],4'b0};
        else
            to_b = breg;                
    end
    
//=========================================================
// OPERAND REGISTERS
//=========================================================
    always @(posedge clk ) begin
        if (rst) begin
            a <= 0;
            b <= 0;
            dout <= 0;
        end
        else begin
            if (pipe_ce) begin
                a <= op_a;
                b[3:0] <= op_b[3:0];
                dout <= breg;
                b16 <= to_b16;
                sra_sign_ext <= to_sra_sign_ext;
            end
            if (b15_4_ce)
                b[15:4] <= op_b[15:4];
        end
    end
    
//=========================================================
// ASYNC MULTIPLIER
//=========================================================
// b16 is 1 only if b is 0, to shift a to the upper 16 bits
    always @ (a or b or b16) begin
        mul = a * {b16,b}; 
    end
    
//=========================================================
// ALU
//=========================================================
    always @(a or b or add or ci) begin
        if (add) begin
            {co0,sum,ign} = {a,ci} + {b,1'b1};
            co = co0;
        end
        else begin
            {co0,sum,ign} = {a,ci} - {b,1'b1};
            co = ~co0;
        end
        
        z = sum == 0;
        n = sum[15];
        // sum[15] = a[15] ^ b[15] ^ c15 <===> c15 = sum[15] ^ a[15] ^ b[15]
        v = co0 ^ sum[15] ^ a[15] ^ b[15];
        
    end
    
    always @(a or b or logicop) begin
        case (logicop)
        0: logic <= a&b;
        1: logic <= a|b;
        2: logic <= a^b;
        3: logic <= a&~b;
        endcase
    end

//=========================================================
// address/PC unit
//=========================================================
    always @(branch or pc or brdispext or brdisp) begin
        if (branch)
            pcnext <= pc + {brdispext[6:0],brdisp[7:0],1'b0};
        else
            pcnext <= pc + 2;
    end
    
    always @ (sum,selpc,pcnext,sync_reset) begin
        if (sync_reset)
            addr_nxt <=0;
        else if (selpc)
            addr_nxt <= pcnext;
        else addr_nxt <= sum;
    end
    
    always @(posedge clk ) begin
        if (rst)
            ret <= 0;
        else if (ret_ce)
            ret <= pc;
    end
  
    always @(posedge clk ) begin
        if (rst)
            addr <= 16'b0;
        else if (rdy)
            addr <= addr_nxt;
    end

//=========================================================
// Result select multiplexer
//=========================================================
    
    always @ (res_sel, sum,logic,mul,ret,shr_res,external) begin
        case (res_sel)
            SEL_EXTERNAL    : res = external;
            SEL_SUM         : res = sum;
            SEL_LOGIC       : res = logic;
            SEL_RIGHT_SH    : res = shr_res;
            SEL_MUL_LEFT_SH : res = mul[15:0];
            SEL_RET         : res = ret;
            default         : res = 16'b0;
        endcase
    end


endmodule

