`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company:         BME-VIK
// Engineer:        Peter Czako
// 
// Create Date:     07/06/2007  
// Design Name: 
// Module Name:     ibrams 
// Project Name:    XSOC
// Target Devices:  SPARTAN-3, SPARTAN-3E
// Tool versions:   XILINX ISE 8.2
// Description:     Top module for the instruction and data RAMs of the XSOC 
//                  system. BRAM macros used as memory instances. 
//                  Initialization fields hold the basic startup program code of
//                  the xr16 processor. Init code can be generated from hex
//                  files by the hex2brams.exe application included in the 
//                  XSOC project.
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////

//=========================================================
// MODULE DECLARATION
//=========================================================

module ibrams(
// XSOC BUS INTERFACE
    input           clk,
    input           rst,                
    input           rnw,
    input   [1:0]   be,
    input   [15:0]  addr, 
    input   [15:0]  din,  
    output  [15:0]  dout
// USER IO PORTS
    );
    
//========================================================= 
// XSOC BUS INTERFACE LOGIC
//========================================================= 

    parameter   BASE_ADDRESS = 16'h0000;

    wire [15:0] to_dout;
    wire        sel;

    assign sel          = addr[15:12] == BASE_ADDRESS >> 12; // 2k*16bit
    assign dout[7:0]    = ~clk & sel & rnw & be[0] ? to_dout[7:0] : 8'b0;
    assign dout[15:8]   = ~clk & sel & rnw & be[1] ? to_dout[15:8] : 8'b0;
    
//========================================================= 
// SUBMODULES
//========================================================= 

// Two 8 bit wide internal Block RAMs connected paralelly (MSB, LSB)

// RAMB16_S9: Virtex-II/II-Pro, Spartan-3 2k x 8 + 1 Parity bit Single-Port RAM
// Xilinx HDL Language Template version 6.3.1i

// MSB BRAM INSTANTIATION
   RAMB16_S9 msb_bram (
                  .DO(to_dout[15:8]),           // 8-bit Data Output
                  .DOP(),                       // 1-bit parity Output
                  .ADDR(addr[11:1]),            // 11-bit Address Input
                  .CLK(~clk),                   // Clock
                  .DI(din[15:8]),               // 8-bit Data Input
                  .DIP(1'b0),                   // 1-bit parity Input
                  .EN(1'b1),                    // RAM Enable Input
                  .SSR(),                       // Synchronous Set/Reset Input
                  .WE(~rnw & sel & be[1])       // Write Enable Input
   );

// LSB BRAM INSTANTIATION
   RAMB16_S9 lsb_bram (
                  .DO(to_dout[7:0]),            // 8-bit Data Output
                  .DOP(),                       // 1-bit parity Output
                  .ADDR(addr[11:1]),            // 11-bit Address Input
                  .CLK(~clk),                   // Clock
                  .DI(din[7:0]),                // 8-bit Data Input
                  .DIP(1'b0),                   // 1-bit parity Input
                  .EN(1'b1),                    // RAM Enable Input
                  .SSR(),                       // Synchronous Set/Reset Input
                  .WE(~rnw & sel & be[0])       // Write Enable Input
   );


// The following defparam declarations are only necessary if you wish to change the default behavior
// of the RAM. If the instance name is changed, these defparams need to be updated accordingly.

   defparam msb_bram.WRITE_MODE = "WRITE_FIRST"; // WRITE_FIRST, READ_FIRST or NO_CHANGE
   defparam lsb_bram.WRITE_MODE = "WRITE_FIRST"; // WRITE_FIRST, READ_FIRST or NO_CHANGE
   
   // The following defparam INIT_xx declarations are only necessary if you wish to change the initial
   // contents of the RAM to anything other than all zero's.

// ------------------------------- PROGRAM INITIALIZATION ------------------------------------------------

                              // ,f,e,d,c,b,a,9,8,7,6,5,4,3,2,1,0.f.e.d.c.b.a.9.8.7.6.5.4.3.2.1.0
//Program init MSB    empty.hex
defparam msb_bram.INIT_00 = 256'h573656D628D08F8E8D8C8B8A898887868584838221D081D0B02127D026D62DD0;
defparam msb_bram.INIT_01 = 256'h00000000000000A0919187A051D05F5E5D5C5B5A595857565554535221D0AF47;
defparam msb_bram.INIT_02 = 256'hA062D6B2104969D600000000000000A0C0C0C0C0C0C0C0C00000000000000000;
defparam msb_bram.INIT_03 = 256'h00A02DD05F5C5BB3106909C063092B09B00B0C8F8C8B2DDFA093D6B3104969D6;
defparam msb_bram.INIT_04 = 256'hD049D0090C09C0B0C023D08F8C8B2DDFA02DD05F99D63029D0C090D690D68F2D;
defparam msb_bram.INIT_05 = 256'h08C0893009C0B320DF49D0093B0B49D0094848D058C0893009C0BB20DFBB1028;
defparam msb_bram.INIT_06 = 256'h5909B320DF49D009B099300908C0B320DF49D009B08730373049D0094747D057;
defparam msb_bram.INIT_07 = 256'h0000A02DD05F5C5BB0B0B320DF49D009B0C06309B0C043D0035909C043D00349;
defparam msb_bram.INIT_08 = 256'h0000000000000000000000000000000000000000000000A02D5CB02CB00C8C2D;


                              // ,f,e,d,c,b,a,9,8,7,6,5,4,3,2,1,0.f.e.d.c.b.a.9.8.7.6.5.4.3.2.1.0
//Program init LSB    empty.hex
defparam lsb_bram.INIT_00 = 256'h6018000000091D1B19171513111E1C1A18161412020702072C01002000000EFF;
defparam lsb_bram.INIT_01 = 256'h0000000000000070666264E002071D1B19171513111E1C1A1816141202077084;
defparam lsb_bram.INIT_02 = 256'hF00205FA9004030500000000000000F00A0A0A0A0A0A0A100000000000000000;
defparam lsb_bram.INIT_03 = 256'h00F0DC00DAD8D6F69090BC0C909C91B0040030DAD8D6D4FFF00005FA90020105;
defparam lsb_bram.INIT_04 = 256'h030F0FC090200B5D0C0B01DAD8D6D0FFF0D800D603050008001203050600D6D8;
defparam lsb_bram.INIT_05 = 256'hB00BDC00200B1190FD0F0FC019800F0F20880F0FDE0BDE00200B499DFC4C8900;
defparam lsb_bram.INIT_06 = 256'h90B00D9EFC0F0FC01E800020B00B059FFC0F0FC02A800019000F0F20880F0FDC;
defparam lsb_bram.INIT_07 = 256'h0000F0D001DAD8D6A000009CFC0F0FC0060C90B00A0C0F0F9090B00C0F0F9078;
defparam lsb_bram.INIT_08 = 256'h0000000000000000000000000000000000000000000000F0D2D0FDC10000D0DE;


// End of BRAM instantiation



endmodule
