`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company:         BME-VIK
// Engineer:        Peter Czako
// 
// Create Date:     20:10:03 01/16/2007 
// Design Name: 
// Module Name:     it_ctrl
// Project Name:    XSOC
// Target Devices:  SPARTAN-3, SPARTAN-3E
// Tool versions:   XILINX ISE 8.2
// Description:     Interrupt controller of the XSOC system. Handles 8 maskable 
//                  interrupt sources with fix priority (0-highest, 7-lowest).
//                  Provides an interrupt vector relative address for the 
//                  handler software.                   
//                  
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////

//=========================================================
// MODULE DECLARATION
//=========================================================

module it_ctrl(
// XSOC BUS INTERFACE
    input           clk,
    input           rst,                
    input           rnw,
    input   [1:0]   be,
    input   [15:0]  addr, 
    input   [15:0]  din,  
    output  [15:0]  dout,
    output          irq,
    input           irq_served,
// USER IO PORTS    
    input       [7:0] irq_i,            // from peripherals 
    output  reg [7:0] irq_served_o      // to peripherals
    );

//========================================================= 
// XSOC BUS INTERFACE LOGIC 
//=========================================================

    parameter   BASE_ADDRESS = 16'h6000;

    reg [15:0]  to_dout;

    wire        sel; 

    assign sel          = addr[15:2] == BASE_ADDRESS >> 2 ;
    assign dout[7:0]    = ~clk & sel & rnw & be[0] ? to_dout[7:0] : 8'b0;
    assign dout[15:8]   = ~clk & sel & rnw & be[1] ? to_dout[15:8] : 8'b0;

//========================================================= 
// USER LOGIC
//========================================================= 

// REGS 
    reg         taken;              // Irq is under service
    reg [7:0]   irq_enable;         // 1-enable, 0-disable current irq
    reg [2:0]   irq_taken_id;       // Number of the currently serviced irq 
    reg [7:0]   irq_status;         // Actual irq input values
    
// WIRES
    wire [7:0]  irq_masked;         // Irq inputs masked with irq_enable
    wire [15:0] vector_addr;        // Interrupt service routine relative address
    
    
// ASSIGNS  
    assign irq_masked   = irq_i & irq_enable; 
    assign vector_addr  = {12'b0,irq_taken_id, 1'b0};

    // To reduce one clock delay of irq acknowledge in the xr16,
    // the incoming irq signals are not registered here,
    // and are directly connected to the processors 
    // input. All interrupt request signals must be registered
    // to negedge clk in the peripherals
    
    assign irq          = |(irq_masked);
    
//========================================================= 
// WRITE REGISTERS
//========================================================= 
    always @(negedge clk) begin
        if (rst)
            irq_enable <=0;
        else if (sel & ~rnw & be[1] & addr[1])
            irq_enable <= din[15:8];
    end

//========================================================= 
// READ REGISTERS
//========================================================= 
    always @(negedge clk) begin
        if (rst)
            to_dout <= 16'h0;
        else begin
            if (sel & rnw & ~addr[1])
                to_dout <= vector_addr; // READ_ONLY
            else if (sel & rnw & addr[1])
                to_dout <= {irq_enable,irq_status};
            else 
                to_dout <= 16'h0;
        end
    end

//========================================================= 
// REGISTERED IRQ INPUT
//========================================================= 
    always @ (negedge clk) begin
        if (rst)
            irq_status <= 0;
        else 
            irq_status <= irq_i;
    end
    
//========================================================= 
// FIX PRIORITY IRQ HANDLING        
//========================================================= 
    always @ (negedge clk) begin
        if (rst) begin
            irq_taken_id <= 0;
            taken <= 0;
        end
        else begin
            if (irq_served) begin
                taken <= 0;
            end
            if (|irq_masked && ~taken) begin 
                taken <= 1;
                if (irq_masked[0])
                    irq_taken_id <= 3'h0;
                else if (irq_masked[1])
                    irq_taken_id <= 3'h1;
                else if (irq_masked[2])
                    irq_taken_id <= 3'h2;
                else if (irq_masked[3])
                    irq_taken_id <= 3'h3;
                else if (irq_masked[4])
                    irq_taken_id <= 3'h4;
                else if (irq_masked[5])
                    irq_taken_id <= 3'h5;
                else if (irq_masked[6])
                    irq_taken_id <= 3'h6;
                else if (irq_masked[7])
                    irq_taken_id <= 3'h7;
            end
        end
    end 

//========================================================= 
// DECODING IRQ SERVED SIGNALS
//========================================================= 
always @ (irq_taken_id,irq_served) begin
    if(irq_served)
        irq_served_o <= 8'h01 << irq_taken_id;
    else
        irq_served_o <= 8'h00;
end
    
endmodule
