`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: BME-VIK
// Engineer:  Peter Czako
// 
// Create Date:     06/27/2007  
// Design Name: 
// Module Name:     segment 
// Project Name:    XSOC
// Target Devices:  SPARTAN-3, SPARTAN-3E
// Tool versions:   XILINX ISE 8.2
// Description:     7 segment display controller module of the XSOC system.
//                  Two modes are available: hexadecimal display OR free 
//                  separate control of the anodes. Working with 1 kHz refresh
//                  frequency for the 4 anodes.
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////

//========================================================= 
// MODULE DECLARATION
//========================================================= 

module segment(
// XSOC BUS INTERFACE
    input           clk,
    input           rst,                
    input           rnw,
    input   [1:0]   be,
    input   [15:0]  addr,    
    input   [15:0]  din, 
    output  [15:0]  dout,
// USER IO PORTS    
    output  [7:0]   seg_out,    // Segment control
    output  [3:0]   an_out      // Anode control
    );
    
//========================================================= 
// XSOC BUS INTERFACE LOGIC 
//========================================================= 
    parameter   BASE_ADDRESS = 16'h6040;

    reg [15:0]  to_dout;

    wire        sel;
    
    assign sel          =  addr[15:3] == BASE_ADDRESS >> 3;    
    assign dout[7:0]    = ~clk & sel & rnw & be[0] ? to_dout[7:0] : 8'b0;
    assign dout[15:8]   = ~clk & sel & rnw & be[1] ? to_dout[15:8] : 8'b0;

//=========================================================     
// USER LOGIC   
//========================================================= 

// PARAMETERS   
    parameter   ANODE_DIV = 12000;  // 12 MHz -> 1 kHz
    
// REGS
    reg         mode;               // 1-control anodes separately, 0-hexadecimal 
    reg [3:0]   hex_data;           // 4 bit data to control 1 anode
    reg [7:0]   anode0;             // Segment control registers for the 4 anodes
    reg [7:0]   anode1;
    reg [7:0]   anode2;
    reg [7:0]   anode3;
    reg [7:0]   sep_segm;           // Segment control for the actual anode (mode = 1)
    reg [13:0]  cnt;                // Display clock generator counter
    reg [15:0]  data;               // 16 bit data in hexadecimal mode
    
// WIRES    
    wire        clk_disp_ce;        // Display clock clock enable
    wire [1:0]  ansel;              // Segment multiplexer select signal
    wire [7:0]  hex_segm;           // Segment control for the actual anode (mode = 0)
    
// ASSIGNS
    assign seg_out      = mode ? sep_segm : hex_segm;
    assign clk_disp_ce  = cnt == ANODE_DIV;

//=========================================================     
// WRITE REGISTERS
//========================================================= 
    always @(negedge clk) begin
        if (rst) begin
            data <= 0;
            anode0  <= 8'hFF;
            anode1  <= 8'hFF;
            anode2  <= 8'hFF;
            anode3  <= 8'hFF;
            mode    <= 0;
        end 
        else if ( sel & ~rnw) begin
            case ({addr[2:1],be})
                4'b00_10 : anode0 <= din[15:8];
                4'b00_01 : anode1 <= din[7:0];
                4'b01_10 : anode2 <= din[15:8];
                4'b01_01 : anode3 <= din[7:0];
                4'b10_01 : data[7:0]  <= din[7:0];
                4'b10_10 : data[15:8] <= din[15:8];
                4'b10_11 : data <= din;
                4'b11_10 : mode <= din[8];
            endcase
        end
    end
    
//=========================================================     
// READ REGISTERS
//========================================================= 
    always @(negedge clk) begin
        if (rst)
            to_dout <= 0;
        else if ( sel & rnw) begin
            case (addr[2:1])
                2'b00 : to_dout <= {anode0,anode1};
                2'b01 : to_dout <= {anode2,anode3};
                3'b10 : to_dout <= data;
                3'b11 : to_dout <= {7'b0,mode,7'b0,mode};
            endcase
        end
        else
            to_dout <= 16'b0;
    end

//========================================================= 
// ANODE CLOCK GENERATION
//=========================================================     
    always @ (negedge clk) begin
        if (rst | cnt == ANODE_DIV) 
            cnt <= 0;
        else 
            cnt <= cnt + 1;
    end

//========================================================= 
// SEGMENT CONTROL SIGNALS
//========================================================= 
    always @ (ansel,data, anode0, anode1, anode2, anode3)
    begin
        case (ansel) 
            2'b00:  begin
                        hex_data <= data[3:0];
                        sep_segm <= anode0;
                    end
            2'b01:  begin
                        hex_data <= data[7:4];
                        sep_segm <= anode1;
                    end
            2'b10:  begin
                        hex_data <= data[11:8];
                        sep_segm <= anode2;
                    end
            2'b11:  begin
                        hex_data <= data[15:12];
                        sep_segm <= anode3;
                    end
        endcase
    end

//=========================================================
// SUBMODULES   
//=========================================================     

// ROM TO CONVERT HEXADECIMAL DATA TO SEGMENT CONTROL SIGNALS
    seg_table st(
                .datain(hex_data),
                .segctrl(hex_segm)
                );
// ANODE SELECT SIGNAL GENERATION   
    anode_sel anode_s(
                .clk(~clk),
                .clk_ce(clk_disp_ce),
                .rst(rst),
                .anode(an_out),
                .an_2bit(ansel)
                );

endmodule
