`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company:         BME-VIK
// Engineer:        Peter Czako
// 
// Create Date:     20:46:38 01/23/2007 
// Design Name: 
// Module Name:     sram_ctrl 
// Project Name:    XSOC
// Target Devices:  SPARTAN-3, SPARTAN-3E
// Tool versions:   XILINX ISE 8.2
// Description:     External SRAM controller. 128k 8 bit wide memory mapped into
//                  4 32k large pages. The pages are mapped to the 0x8000-0xFFFF
//                  address domain. Read operations take 1 clock cycle if the 
//                  half period of the clock is longer than the SRAM's access 
//                  time. Byte write operation takes 2 cycles, word write needs
//                  3 clock cycles.
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////

//=========================================================
// MODULE DECLARATION
//=========================================================

module sram_ctrl(
// XSOC BUS INTERFACE
    input           clk,
    input           rst,                
    input           rnw,
    input   [1:0]   be,
    input   [15:0]  addr,    
    input   [15:0]  din, 
    output  [15:0]  dout,
    output          nwait,              
// USER IO PORTS    
    output  [16:0]  ext_addr,       // External address bus
    inout   [7:0]   ext_data,       // External data bus
    output  reg     cs,             // SRAM Chip Select signal
    output          oe,             // SRAM Output Enable signal
    output          we              // SRAM Write Enable signal
    );
    
//========================================================= 
// XSOC BUS INTERFACE LOGIC 
//=========================================================
    parameter   BASE_ADDRESS = 16'h6060;

    reg [15:0]  to_dout;
    
    wire        sel;                // Select the mapped memory
    wire        mem_ctrl_sel;       // Select the memory controller

    assign sel          = addr[15]; // 4 pages, each from 0x8000 to 0xFFFF  
    assign mem_ctrl_sel = addr == BASE_ADDRESS;
    assign dout[7:0]    = ~clk & (sel | mem_ctrl_sel) & rnw & be[0] ? to_dout[7:0] : 8'b0;
    assign dout[15:8]   = ~clk & (sel | mem_ctrl_sel) & rnw & be[1] ? to_dout[15:8] : 8'b0;
    
//========================================================= 
// USER LOGIC   
//========================================================= 

// REGS
    reg         ext_a0;         // External address 0 signal
    reg         ph12;           // Internal operational phases 
    reg         ph23;           
    reg         ph34;
    reg         ph45;
    reg [1:0]   page_sel;       // Selects the actual page 
    reg [7:0]   msb_din;        // Register for external data
    
// WIRES
    wire        w_cs;           // Write pulses in the cs signal

// ASSIGNS  
    assign ext_addr     = {page_sel,addr[14:1],ext_a0};
    assign ext_data     = ~(sel & ~rnw) ? 8'bz : (ext_a0 ? din[7:0] : din[15:8]);
    assign oe           = ~(sel & rnw);
    assign we           = ~(sel & ~rnw);  
    assign nwait        = (be[1] & be[0]) ? ~(ph12 | ph23 | ph34) : ~ph12;
    assign w_cs         = ~( (ph12 & ~ph23) | (~ph23 & ph34 & be[1] & be[0]) );


//=========================================================
// WRITE REGISTERS
//=========================================================
    always @ (negedge clk) begin
        if (rst)
            page_sel <= 0;
        else if (mem_ctrl_sel & ~rnw & be[1] & ~be[0])
            page_sel <= din[9:8];
    end     

//=========================================================
// READ REGISTERS
//=========================================================

// Combinational output used to perform 1 cycle word read operation
// This is only available if half of clk's period 
// is higher than the SRAM's access time 

    always @* begin
        if (sel & rnw)
            to_dout = {msb_din, ext_data};
        else if (mem_ctrl_sel & rnw)
            to_dout = {6'b0,page_sel,6'b0,page_sel};
        else
            to_dout = 0;
    end 


//=========================================================
// REGISTER EXTERNAL SRAM DATA
//=========================================================

// Read Byte - msb_din == ext_data
// Read Word - MSB read first, then address is increased

    always @ (negedge clk) begin
        if (rst)
            msb_din <= 0;
        else if(sel & rnw)
            msb_din <= ext_data;
    end
    
//========================================================= 
// INTERNAL OPERATIONAL PHASES (ph12,ph23,ph34,ph45)
//=========================================================
 
    always @ (negedge clk) begin
        if (rst | ph12)
            ph12 <= 0;
        else if(sel & ~rnw & ~ph34)
            ph12 <= 1;
    end
    
    always @ (posedge clk) begin
        if (rst)
            ph23 <= 0;
        else if(ph12 & be[1] & be[0])
            ph23 <= 1;
        else 
            ph23 <= 0;
    end
    
    always @ (negedge clk) begin
        if (rst)
            ph34 <= 0;
        else if(ph23)
            ph34 <= 1;
        else 
            ph34 <= 0;
    end
    
    always @ (posedge clk) begin
        if (rst)
            ph45 <= 0;
        else if(ph34)
            ph45 <= 1;
        else 
            ph45 <= 0;
    end

//========================================================= 
// SRAM CONTROL SIGNALS (cs,ext_a0)
//=========================================================


    always @ (sel, rnw, w_cs) begin
        if (sel & rnw)
            cs = ~sel;
        else if (sel & ~rnw)
            cs = w_cs;
        else 
            cs = 1;
    end

    always @ (sel, rnw, be, clk, ph34, ph45,addr[0]) begin
        if (sel & be[1] & be[0] & rnw)
            ext_a0 = ~clk;
        else if (sel & be[1] & be[0] & ~rnw)
            ext_a0 = (ph34 | ph45);
        else 
            ext_a0 = addr[0];
    end

endmodule
