`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company:         BME-VIK
// Engineer:        Peter Czako
// 
// Create Date:     11/25/2006  
// Design Name: 
// Module Name:     xr16 
// Project Name:    XSOC
// Target Devices:  SPARTAN-3, SPARTAN-3E
// Tool versions:   XILINX ISE 8.2
// Description:     Top module for the xr16 processor. Based on the design of
//                  Jan Gray. Modifications listed in the control and the 
//                  datapath submodules. 
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////

/* xr16.v -- xr16 pipelined RISC processor synthesizable Verilog model
 *
 * Copyright (C) 1999, 2000, Gray Research LLC.  All rights reserved.
 * The contents of this file are subject to the XSOC License Agreement;
 * you may not use this file except in compliance with this Agreement.
 * See the LICENSE file.
 *
 * $Header: /dist/xsocv/xr16.v 7     4/06/00 10:55a Jan $
 * $Log: /dist/xsocv/xr16.v $
 * 
 * 7     4/06/00 10:55a Jan
 * polish
 */ 

//=========================================================
// MODULE DECLARATION
//=========================================================

module xr16(
// XSOC BUS INTERFACE
    input           clk,        
    input           rst,        
    output          rnw,
    output   [1:0]  be,
    output  [15:0]  addr,   
    input   [15:0]  din,
    output  [15:0]  dout,
    input           nwait,      
    input           irq,    
    output          irq_served,
    input           dma_req,
    output          dma_ack
    );

//=========================================================
// LOCALS
//=========================================================
    wire            z;                  // zero result condition code
    wire            n;                  // negative result condition code
    wire            co;                 // carry-out result codition code
    wire            v;                  // oVerflow result condition code
    wire            word;               // next access is 16 bit wide
    wire            rf_we;              // register file write enable
    wire    [3:0]   rna;                // register file port A register number
    wire    [3:0]   rnb;                // register file port B register number
    wire            fwd;                // forward result bus into A operand register
    wire    [11:0]  imm;                // 12-bit immediate field
    wire            sextimm4;           // sign-extend 4-bit immediate operand
    wire            zextimm4;           // zero-extend 4-bit immediate operand
    wire            wordimm4;           // word-offset 4-bit immediate operand
    wire            imm12;              // 12-bit immediate operand
    wire            pipe_ce;            // pipeline clock enable
    wire            b15_4_ce;           // b[15:4] clock enable
    wire            add;                // 1 => A + B; 0 => A - B
    wire            ci;                 // carry-in
    wire    [1:0]   logicop;            // logic unit opcode
    wire    [2:0]   res_sel;            // result select signals
    wire            sll;                // shift left logical operand
    wire            srl;                // shift right logical operand
    wire            sra;                // shift right arithmetical operand
    wire            branch;             // branch taken
    wire    [7:0]   brdisp;             // 8-bit branch displacement
    wire            selpc;              // address mux selects next PC
    wire            pc_ce;              // PC clock enable
    wire            ret_ce;             // return address clock enable
    wire            sync_reset;         // 1 cycle delayed reset

    wire    [15:0]  to_dout;            // data out signal before being gated   
    wire    [15:0]  to_addr;            // address signal before being gated   
    wire            to_rnw;             // read /~write before being gated   
    wire    [1:0]   to_be;              // byte enable before being gated   

//=========================================================
// OUTPUT SIGNALS INACTIVE IN CASE OF DMA TRANSFER
//=========================================================
    assign dout[7:0]    = (rnw | dma_ack | ~to_be[0]) ? 8'b0 : to_dout[7:0];
    assign dout[15:8]   = (rnw | dma_ack | ~to_be[1]) ? 8'b0 : to_dout[15:8];
    assign addr         = dma_ack ?  16'b0 : to_addr;
    assign rnw          = dma_ack ?  1'b1  : to_rnw;
    assign be           = dma_ack ?  2'b11 : to_be;
    
//=========================================================
// SUBMODULES
//=========================================================
// CONTROL
    control ctrl_inst(
        .clk(clk),
        .rst(rst),
        .z(z),
        .n(n),
        .co(co),
        .v(v),
        .word(word),
        .rdy(rdy),
        .rf_we(rf_we),
        .rna(rna),
        .rnb(rnb),
        .fwd(fwd),
        .imm(imm),
        .sextimm4(sextimm4),
        .zextimm4(zextimm4),
        .wordimm4(wordimm4),
        .imm12(imm12),
        .pipe_ce(pipe_ce),
        .b15_4_ce(b15_4_ce),
        .add(add),
        .ci(ci),
        .logicop(logicop),
        .res_sel(res_sel),
        .sll(sll),
        .sra(sra),
        .srl(srl),
        .branch(branch),
        .brdisp(brdisp),
        .selpc(selpc),
        .sync_reset(sync_reset),
        .pc_ce(pc_ce),
        .ret_ce(ret_ce),
        
        .nwait(nwait),
        .insn(din),
        .rnw(to_rnw),
        .dma_req(dma_req),
        .dma_ack(dma_ack),
        .irq(irq),
        .irq_served(irq_served)
        );

// DATAPATH
    datapath dp(
        .clk(clk),
        .rst(rst),
        .z(z),
        .n(n),
        .co(co),
        .v(v), 
        .word(word),     
        .rdy(rdy),
        .rf_we(rf_we),
        .rna(rna),
        .rnb(rnb),
        .fwd(fwd),
        .imm(imm),
        .sextimm4(sextimm4),
        .zextimm4(zextimm4),
        .wordimm4(wordimm4),
        .imm12(imm12),
        .pipe_ce(pipe_ce),
        .b15_4_ce(b15_4_ce),
        .add(add),
        .ci(ci),
        .logicop(logicop), 
        .res_sel(res_sel),
        .sll(sll),
        .sra(sra),
        .srl(srl),
        .branch(branch),
        .brdisp(brdisp),
        .selpc(selpc),
        .sync_reset(sync_reset),
        .pc_ce(pc_ce),
        .ret_ce(ret_ce),
        
        .data_in(din),
        .addr(to_addr),
        .data_out(to_dout),
        .be(to_be));



endmodule
