`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company:         BME-VIK
// Engineer:        Peter Czako
// 
// Create Date:     11/25/2006  
// Design Name: 
// Module Name:     xsoc 
// Project Name:    XSOC
// Target Devices:  SPARTAN-3, SPARTAN-3E
// Tool versions:   XILINX ISE 8.2
// Description:     Top module for the XSOC system. Declares external input
//                  and output signals, and establishes connectivity of the 
//                  submodules. Prepared to be used with SPARTAN-3 Starter
//                  Board and SPARTAN-3E Small Board.
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: Used for educational purposes only
//
//////////////////////////////////////////////////////////////////////////////////

/* xsoc.v -- XSOC System-on-a-Chip synthesizable Verilog model
 *
 * Copyright (C) 1999, 2000, Gray Research LLC.  All rights reserved.
 * The contents of this file are subject to the XSOC License Agreement;
 * you may not use this file except in compliance with this Agreement.
 * See the LICENSE file.
 *
 * $Header: /dist/xsocv/xsoc.v 8     4/06/00 10:55a Jan $
 * $Log: /dist/xsocv/xsoc.v $
 * 
 * 8     4/06/00 10:55a Jan
 * polish
 */ 

// IF USED WITH SPARTAN-3E SMALL BOARD
// UNCOMMENT THE FOLLOWING LINE
//`define SP3E_SMALL        

//=========================================================
// MODULE DECLARATION
//=========================================================
module xsoc(
    input           clk_in,         // global clock
    input           rst_btn,        // global reset button
//  input           rst_logsys,     // global reset through logsys cable
    input   [7:0]   switch,         // Switches
    input   [2:0]   button,         // Pushbuttons (BTN0-BTN3)
    input           uart_rx,        // Serial uart receive 
    output          uart_tx,        // Serial transmit
    output  [7:0]   led,            // LEDs
    output  [7:0]   segment,        // 7 Segment display segment control
    output  [3:0]   anode,          // 7 Segment display anode control

// DIFFERENT EXTERNAL SRAM INTERFACE
// FOR SPARTAN-3 STARTER BOARD AND SPARTAN-3E SMALL BOARD

`ifdef SP3E_SMALL                   //      128k * 8
    output  [16:0]  sram_addr,      // External SRAM address bus
`else       // SPARTAN-3 STARTER BOARD      2 * 256k * 16
    output  [17:0]  sram_addr,
    output          lb,             // External SRAM lower byte enable
    output          ub,             // External SRAM upper byte enable
`endif  
    output          cs,             // External SRAM chip select
    output          oe,             // External SRAM output enable
    output          we,             // External SRAM write enable
    inout   [7:0]   sram_data       // External SRAM high impedance data bus
    );

//=========================================================
// PARAMETERS
//=========================================================
// DIFFERENT INPUT CLOCK 
// FOR SPARTAN-3 STARTER BOARD AND SPARTAN-3E SMALL BOARD

`ifdef SP3E_SMALL                        // 16 MHz -> 12 MHz
    parameter CLK_M = 3;
    parameter CLK_D = 4;
`else       // SPARTAN-3 STARTER BOARD      50 MHz -> 12 MHz
    parameter CLK_M = 6;
    parameter CLK_D = 25;
`endif      

//=========================================================
// WIRES
//=========================================================

// XSOC BUS WIRES
    wire            clk;                    // 12 MHz system clock
    wire            rst;                    // Reset signal
    wire            rnw;                    // Read/~Write
    wire    [1:0]   be;                     // Byte enable be[1] => MSB, be[0] => LSB   
    wire    [15:0]  addr;                   // 16 bit address bus 
    wire    [15:0]  din;                    // 16 bit input data bus
    wire            nwait;                  // Wait request signal, stops the processor 
    wire            irq;                    // Interrupt controller interrupt request towards the xr16
    wire            irq_served;             // Interrupt controller interrupt served from the xr16
    wire            dma_ack;                // DMA acknowledge
    wire            dma_req;                // DMA request 
    
// SEPARATE CONTROL SIGNALS FOR BUS MASTERS (XR16 AND DMA CONTROLLER)   

// MASTER 0
    wire            rnw_proc;               // xr16 read/~write
    wire    [1:0]   be_proc;                // xr16 byte enable
    wire    [15:0]  addr_proc;              // xr16 address bus 
    
// MASTER 1
    wire            rnw_dma;                // DMA controller read/~write
    wire    [1:0]   be_dma;                 // DMA controller byte enable
    wire    [15:0]  addr_dma;               // DMA controller address bus
    
// DATA OUT SIGNALS FOR EVERY SUBMODULE

    wire    [15:0]  dout_bp;                // Breakpoint logic output
    wire    [15:0]  dout_itc;               // Interrupt controller output
    wire    [15:0]  dout_led;               // Led controller output
    wire    [15:0]  dout_mem;               // Instruction and data memory output
    wire    [15:0]  dout_proc;              // xr16 processor output
    wire    [15:0]  dout_segm;              // 7 segment display controller output
    wire    [15:0]  dout_sram;              // External SRAM controller output
    wire    [15:0]  dout_sw;                // Switch output
    wire    [15:0]  dout_timer;             // 16 bit timer output
    wire    [15:0]  dout_uart;              // Serial UART output
    
// IRQ SIGNALS

    wire            bp_irq;                 // Breakpoint interrupt request
    wire            uart_irq;               // UART interrupt for signaling ESC characters
    wire            timer_irq;              // Timer interrupt request
    wire    [7:0]   irq_itc_i;              // 8 different interrupt sources of the interrupt controller
    wire    [7:0]   irq_served_itc_o;       // Interrupt served signals of the interrupt controller
    
// SRAM SIGNAL

    wire    [16:0]  ext_addr;               // SRAM controller external address bus         

//=========================================================
// ASSIGNS
//=========================================================

    assign rst  = rst_btn; // | rst_logsys;

// DIFFERENT EXTERNAL SRAM INTERFACE
// FOR SPARTAN-3 STARTER BOARD AND SPARTAN-3E SMALL BOARD

`ifdef SP3E_SMALL               //      WHOLE MEMORY USED (128k * 8)
    assign sram_addr    = {ext_addr};
`else   // SPARTAN-3 STARTER BOARD      LOWER BYTE OF ONE MEMORY USED
    assign sram_addr    = {1'b0, ext_addr};
    assign lb           = 0;
    assign ub           = 1;
`endif  
    
// DMA CONTROLLER IS NOT YET IMPLEMENTED => SIGNALS INACTIVE

    assign be_dma       = 2'b11;
    assign rnw_dma      = 1'b1;
    assign addr_dma     = 16'h0000;
    assign dma_req      = 1'b0;
    
// INTERRUPT REQUEST SIGNALS
// FIX PRIORITY (0-highest, 7-lowest)

// irq_0 serves higher level control purposes
    assign irq_itc_i[0] = bp_irq | uart_irq;        
    assign irq_itc_i[1] = button[2];
    assign irq_itc_i[2] = button[1];
    assign irq_itc_i[3] = button[0];
    assign irq_itc_i[4] = timer_irq;
    assign irq_itc_i[5] = 1'b0;
    assign irq_itc_i[6] = 1'b0;
    assign irq_itc_i[7] = 1'b0;
    
//=========================================================
// SUBMODULES
//=========================================================
    
// BUS GATE MODULE
    bus_gate busgate (
            .rnw_m0_i(rnw_proc), 
            .be_m0_i(be_proc), 
            .addr_m0_i(addr_proc), 
            .rnw_m1_i(rnw_dma), 
            .be_m1_i(be_dma), 
            .addr_m1_i(addr_dma), 
            .dout0_i(dout_bp), 
            .dout1_i(dout_itc), 
            .dout2_i(dout_led), 
            .dout3_i(dout_mem), 
            .dout4_i(dout_proc), 
            .dout5_i(dout_segm), 
            .dout6_i(dout_sram), 
            .dout7_i(dout_sw), 
            .dout8_i(dout_timer), 
            .dout9_i(dout_uart), 
            .dout10_i(16'h0000), 
            .dout11_i(16'h0000), 
            .rnw_o(rnw), 
            .be_o(be), 
            .addr_o(addr), 
            .din_o(din)
            );
    
    
// CLOCK DIVIDER DCM
    clk_div #(
        .DIV(CLK_D),
        .MUL(CLK_M)
        )
        cdiv(
            .clk_i(clk_in),
            .clk_o(clk)
            );


// XR16 PROCESSOR
    xr16 p(
            .clk(clk), 
            .rst(rst),
            .rnw(rnw_proc),
            .be(be_proc),
            .addr(addr_proc),
            .din(din),
            .dout(dout_proc),
            .nwait(nwait),
            .irq(irq),
            .irq_served(irq_served),
            .dma_req(dma_req),
            .dma_ack(dma_ack)
            );

// INSTRUCTION AND DATA BRAMS (0x0000 - 0x0FFF)
    ibrams #(.BASE_ADDRESS(16'h0000))
        ibrams_inst(
            .clk(clk),
            .rst(rst),
            .rnw(rnw),
            .be(be),
            .addr(addr),
            .din(din),
            .dout(dout_mem)
            );
            
            
// INTERRUPT CONTROLLER
    it_ctrl #(.BASE_ADDRESS(16'h6000))
        itc(
            .clk(clk), 
            .rst(rst), 
            .rnw(rnw),
            .be(be),
            .addr(addr), 
            .din(din), 
            .dout(dout_itc), 
            .irq(irq),
            .irq_served(irq_served),
            .irq_i(irq_itc_i),
            .irq_served_o(irq_served_itc_o)
            );

// BREAK POINT LOGIC
    bp_logic #(.BASE_ADDRESS(16'h6004))
        bplogic(
            .clk(clk),      
            .rst(rst),      
            .rnw(rnw),
            .be(be),
            .addr(addr),    
            .din(din),
            .dout(dout_bp),
            .irq(bp_irq),
            .irq_served(irq_served_itc_o[0])
            );

// SWITCH CONTROLLER (8 BIT GENERAL PURPOSE INPUT)
    xsoc_in8 #(.BASE_ADDRESS(16'h6020))
        switches(
            .clk(clk), 
            .rst(rst),
            .rnw(rnw),
            .be(be),
            .addr(addr),
            .din(din),
            .dout(dout_sw),
            .in8(switch)
            );

// LED CONTROLLER (8 BIT GENERAL PURPOSE OUTPUT)
    xsoc_out8 #(.BASE_ADDRESS(16'h6030))
        leds(
            .clk(clk), 
            .rst(rst),
            .rnw(rnw),
            .be(be),
            .addr(addr),
            .din(din),
            .dout(dout_led),
            .out8(led)
            );

// 7 SEGMENT DISPLAY 
    segment #(.BASE_ADDRESS(16'h6040))
        segm(
            .clk(clk),  
            .rst(rst), 
            .rnw(rnw),
            .be(be),
            .addr(addr), 
            .din(din),
            .dout(dout_segm),
            .seg_out(segment), 
            .an_out(anode)
            );
        
// UART
    xsoc_uart #(.BASE_ADDRESS(16'h6050))
        uart(
            .clk(clk),  
            .rst(rst), 
            .rnw(rnw),
            .be(be),
            .addr(addr), 
            .din(din),
            .dout(dout_uart),
            .irq(uart_irq),
            .irq_served(irq_served_itc_o[0]),
            .serial_in(uart_rx),
            .serial_out(uart_tx)
            );


// SRAM CONTROLLER
    sram_ctrl #(.BASE_ADDRESS(16'h6060))
        sramctrl(
            .clk(clk), 
            .rst(rst),  
            .rnw(rnw), 
            .be(be),
            .addr(addr), 
            .din(din), 
            .dout(dout_sram), 
            .nwait(nwait), 
            .ext_addr(ext_addr), 
            .ext_data(sram_data), 
            .cs(cs), 
            .oe(oe), 
            .we(we)
        );
        
// 16 BIT TIMER COUNTER
    xsoc_timer16 #(.BASE_ADDRESS(16'h6070))
        xtimer16(
            .clk(clk),  
            .rst(rst), 
            .rnw(rnw),
            .be(be),
            .addr(addr), 
            .din(din),
            .dout(dout_timer),
            .irq(timer_irq),
            .irq_served(irq_served_itc_o[4])
            );

endmodule 
