`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company:         BME-VIK
// Engineer:        Peter Czako
// 
// Create Date:     20:26:32 04/23/2007  
// Design Name: 
// Module Name:     xsoc_timer16 
// Project Name:    XSOC
// Target Devices:  SPARTAN-3, SPARTAN-3E
// Tool versions:   XILINX ISE 8.2
// Description:     16 bit timer counter unit in the XSOC system.
//                  Contains an adjustable prescale divider, capable of 
//                  clear on compare match, and interrupt request.
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////

//========================================================= 
// MODULE DECLARATION
//=========================================================

module xsoc_timer16(
// XSOC BUS INTERFACE
    input           clk,
    input           rst,
    input           rnw,
    input   [1:0]   be,
    input   [15:0]  addr,
    input   [15:0]  din,
    output  [15:0]  dout,
    output  reg     irq,
    input           irq_served
// USER IO PORTS    
    );

//========================================================= 
// XSOC BUS INTERFACE LOGIC 
//=========================================================
    parameter   BASE_ADDRESS = 16'h6070;
    
    reg [15:0]  to_dout;
    
    wire        sel;

    assign sel        = addr[15:3] == BASE_ADDRESS >> 3;     
    assign dout[7:0]  = ~clk & sel & rnw & be[0] ? to_dout[7:0] : 8'b0;
    assign dout[15:8] = ~clk & sel & rnw & be[1] ? to_dout[15:8] : 8'b0;

//=========================================================
// USER LOGIC
//=========================================================

// REGS
    reg         irq_en;             // Interrupt enable
    reg         mode;               // 1-clear on compare, 0-continue
    reg         cnt_ld;             // Counter load
    reg         cnt_ce;             // Counter clock enable
    reg [1:0]   match_rise;         // Samling match signal 
    reg [2:0]   pre_sel;            // Prescaler select
    reg [9:0]   pre_cnt;            // Prescaler counter
    reg [15:0]  match_reg;          // Desired match value
    reg [15:0]  cnt_reg;            // Counter register (access from the bus)
    reg [15:0]  cnt;                // Counter
    
// WIRES
    wire        match;              // Counter was equal to match_reg
    wire        match_pulse;        // Rising edge of match
    
// ASSIGNS  
    assign match        = irq_en & (cnt == match_reg);
    assign match_pulse  = match_rise == 2'b01;

    
//=========================================================
// WRITE REGISTERS
//=========================================================
    always @(negedge clk) begin
        if (rst) begin 
            match_reg <= 0;
            cnt_reg <= 0;
            irq_en <= 0;
            mode <= 0;
            pre_sel <= 0;
        end
        else if (sel & ~rnw & addr[2:1] == 0) begin
            if (be[0])
                match_reg[7:0] <= din[7:0];
            if (be[1])
                match_reg[15:8] <= din[15:8];
        end
        else if (sel & ~rnw & addr[2:1] == 1) begin
            if (be[0])
                cnt_reg[7:0] <= din[7:0];
            if (be[1])
                cnt_reg[15:8] <= din[15:8];
        end
        else if (sel & ~rnw & be[1] & addr[2:1] == 2)
            {irq_en,mode,pre_sel} <= din[12:8];
    end
    
//=========================================================
// READ REGISTERS
//=========================================================
    always @(negedge clk) begin
        if (rst)
            to_dout <= 0;
        else if (sel & rnw & addr[2:1] == 0) 
            to_dout <= match_reg;
        else if (sel & rnw & addr[2:1] == 1)
            to_dout <= cnt;
        else if (sel & rnw & addr[2:1] == 2)
            to_dout <= {3'b0,irq_en,mode,pre_sel,8'b0};
        else
            to_dout <= 0;
    end

//=========================================================
// COUNTER LOAD ENABLE
//=========================================================
    always @(negedge clk) begin
        if (rst)
            cnt_ld <= 0;
        else if (sel & ~rnw & addr[2:1] == 1)
            cnt_ld <= 1;
        else
            cnt_ld <= 0;
    end

//=========================================================
// INTERRUPT REQUEST HANDLING
//=========================================================
    always @ (negedge clk) begin
        if (rst | irq_served)
            irq <= 0;
        else if (match_pulse)
            irq <= 1;
    end

//=========================================================
// PRESCALER
//=========================================================
    always @ (negedge clk) begin
        if(rst)
            pre_cnt <= 0;
        else
            pre_cnt <= pre_cnt + 1;
    end

    always @ (pre_cnt,pre_sel)
        begin
            case(pre_sel)
                0: cnt_ce = 1'b0;           // timer off
                1: cnt_ce = 1'b1;           // /1
                2: cnt_ce = pre_cnt[0];     // /2
                3: cnt_ce = &pre_cnt[2:0];  // /8
                4: cnt_ce = &pre_cnt[4:0];  // /32
                5: cnt_ce = &pre_cnt[6:0];  // /128
                6: cnt_ce = &pre_cnt[8:0];  // /512
                7: cnt_ce = &pre_cnt[9:0];  // /1024
            endcase
        end

//=========================================================
// COUNTER
//=========================================================
    always @ (negedge clk) begin
        if(rst | (match & mode))           // clear on compare match
            cnt <= 0;
        else if (cnt_ld)
            cnt <= cnt_reg;
        else if (cnt_ce)
            cnt <= cnt + 1;
    end

//=========================================================
// MATCH SIGNAL SAMPLING
//========================================================= 
    always @ (negedge clk) begin
        if (rst)
            match_rise <= 0;
        else begin
            match_rise[0] <= match;
            match_rise[1] <= match_rise[0];
        end
    end
    
    
endmodule
