`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company:         BME-VIK
// Engineer:        Peter Czako
// 
// Create Date:     17:37:43 03/05/2007 
// Design Name: 
// Module Name:     xsoc_uart 
// Project Name:    XSOC
// Target Devices:  SPARTAN-3, SPARTAN-3E
// Tool versions:   XILINX ISE 8.2
// Description:     Top module for the serial receiver/transmitter in the 
//                  XSOC system. Contains the KCPSM UART VHDL macros, 
//                  and XSOC bus interface logic.
//                  Currently working on 57600 8N1 mode
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////

//=========================================================
// MODULE DECLARATION
//=========================================================

module xsoc_uart(
// XSOC BUS INTERFACE
    input           clk,
    input           rst,
    input           rnw,
    input   [1:0]   be,
    input   [15:0]  addr,
    input   [15:0]  din,
    output  [15:0]  dout,
    output  reg     irq,
    input           irq_served,
// USER IO PORTS    
    input           serial_in,          // Serial input  (RX)
    output          serial_out          // Serial output (TX)
    );

//=========================================================
// XSOC BUS INTERFACE LOGIC 
//=========================================================
    parameter   BASE_ADDRESS = 16'h6050;
    
    reg [15:0]  to_dout;
    
    wire        sel; 

    assign sel        = addr[15:2] == BASE_ADDRESS >> 2 ;
    assign dout[7:0]  = ~clk & sel & rnw & be[0] ? to_dout[7:0] : 8'b0;
    assign dout[15:8] = ~clk & sel & rnw & be[1] ? to_dout[15:8] : 8'b0;

//========================================================= 
// USER LOGIC
//=========================================================
    // BASE = tx_data
    // +1   = tx_status => 6'b0, tx_buf_full, tx_buf_half
    // +2   = rx_data
    // +3   = rx_status => 4'b0, irq_en, rx_buf_data_pres,rx_buf_full,rx_buf_half
    
// REGS 
    reg         irq_en;             // Interrupt enable
    reg  [7:0]  tx_data;            // Last sent byte
    
// WIRES    
    wire        baud;               // Baud clk for the serial interface
    wire        tx_buf_full;        // Status signals
    wire        tx_buf_half;
    wire        rx_buf_full;
    wire        rx_buf_half;
    wire        rx_buf_data_pres;
    wire        esc_received;       // ESC character arrived
    wire [1:0]  tx_status;          // Concatenated TX status signals
    wire [3:0]  rx_status;          // Concatenated RX status signals
    wire [7:0]  rx_data;            // Last received byte
    
// ASSIGNS
    assign tx_status = {tx_buf_full, tx_buf_half};
    assign rx_status = {irq_en,rx_buf_data_pres,rx_buf_full,rx_buf_half};
    
//========================================================= 
// WRITE REGISTERS
//=========================================================
    always @ (negedge clk) begin
        if (rst) begin
            tx_data <= 0;
            irq_en <= 0;
        end
        else if (sel & ~rnw & ~addr[1] & be[1] & ~be[0])
            tx_data <= din[15:8];
        else if (sel & ~rnw & addr[1] & ~be[1] & be[0])
            irq_en <= din[3];
    end

//=========================================================
// READ REGISTERS
//=========================================================

// The UART modules already give a registered output at negedge clk
    always @* begin
        if (rst)
            to_dout <= 0;
        else if(sel & rnw & ~addr[1]) 
            to_dout <= {tx_data,6'b0,tx_status};
        else if (sel & rnw & addr[1]) 
            to_dout <= {rx_data,4'b0,rx_status};
        else
            to_dout <= 0;
    end

//========================================================= 
// INTERRUPT REQUEST HANDLING
//========================================================= 
    always @ (negedge clk) begin
        if (rst)
            irq <= 0;
        else if (irq_en & esc_received)
            irq <= 1;
        else if (irq_served)
            irq <= 0;
    end

//========================================================= 
// SUBMODULES
//========================================================= 
    
// BAUD CLOCK GENERATION
    baud_clk baudclk (
                    .clk(~clk),
                    .rst(rst),
                    .baud(baud)
                    );
    
// MACRO MODULES OF UART_TX AND UART_RX

    uart_tx uarttx (
                    .data_in(din[15:8]),
                    .write_buffer(sel & ~rnw & ~addr[1] & be[1] & ~be[0]),
                    .reset_buffer(rst),
                    .en_16_x_baud(baud),
                    .serial_out(serial_out),
                    .buffer_full(tx_buf_full),
                    .buffer_half_full(tx_buf_half),
                    .clk(~clk)
                    );
                         
    uart_rx uartrx (
                    .data_out(rx_data),
                    .read_buffer(sel & rnw & addr[1] & be[1] & ~be[0]),
                    .reset_buffer(rst),
                    .en_16_x_baud(baud),
                    .buffer_data_present(rx_buf_data_pres),
                    .serial_in(serial_in),
                    .buffer_full(rx_buf_full),
                    .buffer_half_full(rx_buf_half),
                    .esc_received(esc_received),
                    .clk(~clk)
                    );

endmodule
